"""Format typehints elegantly and and fix automatically created links.

The Sphinx extension :mod:`sphinx_autodoc_typehints` adds type annotations to functions.
This extension modifies the created type annotations in four ways:

#. It formats the annotations more simply and in line with e.g. :mod:`numpy`.
#. It defines a configuration value ``qualname_overrides`` for ``conf.py``
   that overrides automatically created links. It is used like this::

       qualname_overrides = {
           "pandas.core.frame.DataFrame": "pandas.DataFrame",  # fix qualname
           "numpy.int64": ("py:data", "numpy.int64"),  # fix role
           "pandas.api.typing.aliases.AnyArrayLike": ("doc", "pandas:reference/aliases"),
           ...,
       }

   The defaults include :class:`anndata.AnnData`, :class:`pandas.DataFrame`,
   :class:`scipy.sparse.spmatrix` and other classes in :mod:`scipy.sparse`.

   It is necessary since :attr:`~definition.__qualname__` does not necessarily match
   the documented location of the function/class.

   Once either `sphinx issue 4826`_ or `sphinx-autodoc-typehints issue 38`_ are fixed,
   this part of the functionality will no longer be necessary.
#. The config value ``annotate_defaults`` (default: :data:`True`) controls if rST code
   like ``(default: `42`)`` is added after the type.
   It sets sphinx-autodoc-typehints’s option ``typehints_defaults`` to ``'braces'``
#. Type annotations for :class:`tuple` return types are added::

       def x() -> Tuple[int, float]:
           \"""
           Returns:
               a: An integer
               b: A floating point number
           \"""

   will render as:

       :Returns: a : :class:`int`
                     An integer
                 b : :class:`float`
                     A floating point number


.. _sphinx issue 4826: https://github.com/sphinx-doc/sphinx/issues/4826
.. _sphinx-autodoc-typehints issue 38: https://github.com/tox-dev/sphinx-autodoc-typehints/issues/38

"""  # noqa: D300, E501

from __future__ import annotations

from typing import TYPE_CHECKING, cast
from pathlib import Path
from collections import ChainMap
from dataclasses import dataclass

from scanpydoc import metadata, _setup_sig
from scanpydoc.elegant_typehints._role_mapping import RoleMapping

from .example import (
    example_func_prose,
    example_func_tuple,
    example_func_anonymous_tuple,
)


if TYPE_CHECKING:
    from typing import Any
    from collections.abc import Callable

    from sphinx.config import Config
    from docutils.nodes import TextElement, reference
    from sphinx.addnodes import pending_xref
    from sphinx.application import Sphinx
    from sphinx.environment import BuildEnvironment


__all__ = [
    "example_func_anonymous_tuple",
    "example_func_prose",
    "example_func_tuple",
    "setup",
]


HERE = Path(__file__).parent.resolve()

qualname_overrides_default = {
    "anndata.base.AnnData": "anndata.AnnData",
    "anndata.core.anndata.AnnData": "anndata.AnnData",
    "anndata._core.anndata.AnnData": "anndata.AnnData",
    "matplotlib.axes._axes.Axes": "matplotlib.axes.Axes",
    "pandas.core.frame.DataFrame": "pandas.DataFrame",
    "pandas.core.indexes.base.Index": "pandas.Index",
    "scipy.sparse.base.spmatrix": "scipy.sparse.spmatrix",
    "scipy.sparse.csr.csr_matrix": "scipy.sparse.csr_matrix",
    "scipy.sparse.csc.csc_matrix": "scipy.sparse.csc_matrix",
}
qualname_overrides = ChainMap(
    RoleMapping(),
    RoleMapping.from_user(qualname_overrides_default),  # type: ignore[arg-type]
)


def _init_vars(_app: Sphinx, config: Config) -> None:
    cast("RoleMapping", qualname_overrides.maps[0]).update_user(
        config.qualname_overrides
    )
    if (
        "sphinx_autodoc_typehints" in config.extensions
        and config.typehints_defaults is None
        and config.annotate_defaults
    ):
        # override default for “typehints_defaults”
        config.typehints_defaults = "braces"


@dataclass
class PickleableCallable:
    func: Callable[..., Any]

    __call__ = property(lambda self: self.func)


# https://www.sphinx-doc.org/en/master/extdev/event_callbacks.html#event-missing-reference
def _last_resolve(
    app: Sphinx,
    env: BuildEnvironment,
    node: pending_xref,
    contnode: TextElement,
) -> reference | None:
    if "sphinx.ext.intersphinx" not in app.extensions:
        return None

    from sphinx.ext.intersphinx import resolve_reference_detect_inventory

    if (
        ref := qualname_overrides.get(
            (f"{node['refdomain']}:{node['reftype']}", node["reftarget"])
        )
    ) is None:
        return None
    role, node["reftarget"] = ref
    if role is not None:
        node["refdomain"], node["reftype"] = role.split(":", 1)
    return resolve_reference_detect_inventory(env, node, contnode)


@_setup_sig
def setup(app: Sphinx) -> dict[str, Any]:
    """Patches :mod:`sphinx_autodoc_typehints` for a more elegant display."""
    if "sphinx.ext.autodoc" not in app.extensions:
        msg = "`scanpydoc.elegant_typehints` requires `sphinx.ext.autodoc`."
        raise RuntimeError(msg)

    app.add_config_value("qualname_overrides", default={}, rebuild="html")
    app.add_config_value("annotate_defaults", default=True, rebuild="html")
    app.connect("config-inited", _init_vars)
    # Add 1 to priority to run after sphinx.ext.intersphinx
    app.connect("missing-reference", _last_resolve, priority=501)

    from ._formatting import typehints_formatter

    app.config["typehints_formatter"] = PickleableCallable(typehints_formatter)

    from . import _return_tuple, _autodoc_patch

    _autodoc_patch.setup(app)
    _return_tuple.setup(app)

    return metadata



from __future__ import annotations

from types import GenericAlias
from typing import TYPE_CHECKING, TypeAliasType, cast, get_args, get_origin

from sphinx_autodoc_typehints import format_annotation

from scanpydoc import elegant_typehints
from scanpydoc._types import _GenericAlias


if TYPE_CHECKING:
    from typing import Any
    from collections.abc import Sequence

    from sphinx.config import Config


def typehints_formatter(annotation: object, config: Config) -> str | None:
    """Generate reStructuredText containing links to the types.

    Can be used as ``typehints_formatter`` for :mod:`sphinx_autodoc_typehints`,
    to respect the ``qualname_overrides`` option.

    Arguments
    ---------
    annotation
        A type or class used as type annotation.
    config
        Sphinx config containing ``sphinx-autodoc-typehints``’s options.

    Returns
    -------
    reStructuredText describing the type
    """
    if isinstance(annotation, TypeAliasType):
        return format_annotation(annotation.__value__, config)

    if isinstance(annotation, type) and annotation.__module__ == "builtins":
        return None

    if isinstance(annotation, GenericAlias | _GenericAlias):
        args = get_args(annotation)
        annotation = cast("type", get_origin(annotation))
    else:
        args = None
    annotation_cls = annotation if isinstance(annotation, type) else type(annotation)
    if annotation_cls.__module__ in {"typing", "types"}:
        return None

    if isinstance(annotation, type):
        return _fmt_type(annotation, args, config)

    return None  # pragma: no cover


def _fmt_type(cls: type, args: Sequence[Any] | None, config: Config) -> str | None:
    full_name = f"{cls.__module__}.{cls.__qualname__}"
    if (
        override := elegant_typehints.qualname_overrides.get((None, full_name))
    ) is None:
        return None

    role, qualname = override
    if role == "doc":
        label = full_name if config.typehints_fully_qualified else cls.__qualname__
        return f":{role}:`{label} <{qualname}>`"

    tilde = "" if config.typehints_fully_qualified else "~"
    if role is None:
        role = "py:exc" if issubclass(cls, BaseException) else "py:class"
    if args is None:
        formatted_args = ""
    else:
        formatted_args = ", ".join(format_annotation(arg, config) for arg in args)
        formatted_args = rf"\ \[{formatted_args}]"
    return f":{role}:`{tilde}{qualname}`{formatted_args}"



"""Custom pytest fixtures and setup."""

from __future__ import annotations

import sys
import linecache
import importlib.util
from uuid import uuid4
from typing import TYPE_CHECKING
from textwrap import dedent

import pytest


if TYPE_CHECKING:
    from types import ModuleType
    from typing import Any, NamedTuple
    from pathlib import Path
    from collections.abc import Callable, Generator

    from sphinx.testing.util import SphinxTestApp

    from scanpydoc.testing import MakeApp

    class _AppParams(NamedTuple):
        args: tuple[Any, ...]
        kwargs: dict[str, Any]


@pytest.fixture
def make_app_setup(
    tmp_path: Path, make_app: type[SphinxTestApp], app_params: _AppParams
) -> MakeApp:
    def get_builder(buildername: str = "html", **_: object) -> str:
        return buildername  # make sure there are no conflicts

    def make_app_setup(
        buildername: str | None = None,
        /,
        *,
        exception_on_warning: bool = False,
        **conf: Any,  # noqa: ANN401
    ) -> SphinxTestApp:
        (tmp_path / "conf.py").write_text("")

        if buildername is None:
            buildername = get_builder(*app_params.args, **app_params.kwargs)
            app_params.kwargs.pop("buildername", None)

        confoverrides = app_params.kwargs.get("confoverrides", {}).copy()
        confoverrides.update(conf)
        confoverrides.setdefault("suppress_warnings", []).append("app.add_node")

        return make_app(
            buildername=buildername,
            srcdir=tmp_path,
            confoverrides=confoverrides,
            warningiserror=exception_on_warning,
            exception_on_warning=exception_on_warning,
        )

    return make_app_setup


@pytest.fixture
def make_module(
    tmp_path: Path,
) -> Generator[Callable[[str, str], ModuleType], None, None]:
    added_modules = []

    def make_module(name: str, code: str) -> ModuleType:
        code = dedent(code)
        assert name not in sys.modules
        spec = importlib.util.spec_from_loader(name, loader=None)
        assert spec is not None
        mod = sys.modules[name] = importlib.util.module_from_spec(spec)
        path = tmp_path / f"{name}_{str(uuid4()).replace('-', '_')}.py"
        path.write_text(code)
        mod.__file__ = str(path)
        exec(code, mod.__dict__)  # noqa: S102
        linecache.updatecache(str(path), mod.__dict__)
        added_modules.append(name)
        return mod

    yield make_module

    for name in added_modules:
        del sys.modules[name]



from __future__ import annotations

from typing import TYPE_CHECKING
from functools import wraps

from sphinx.ext.autodoc import ClassDocumenter

from . import qualname_overrides


if TYPE_CHECKING:
    from sphinx.application import Sphinx
    from docutils.statemachine import StringList


orig = ClassDocumenter.add_directive_header


@wraps(orig)
def add_directive_header(self: ClassDocumenter, sig: str) -> None:
    orig(self, sig)
    lines = self.directive.result
    inferred_role, direc = (
        ("py:exc", "py:exception")
        if isinstance(self.object, type) and issubclass(self.object, BaseException)
        else ("py:class", "py:class")
    )
    for (old_role, old_name), (new_role, new_name) in qualname_overrides.items():
        role = inferred_role if new_role is None else new_role
        # Currently, autodoc doesn’t link to bases using :exc:
        lines.replace(
            f":{old_role or 'py:class'}:`{old_name}`", f":{role}:`{new_name}`"
        )
        # But maybe in the future it will
        lines.replace(f":{role}:`{old_name}`", f":{role}:`{new_name}`")
        if any("." not in name for name in (old_name, new_name)):
            continue  # pragma: no cover
        old_mod, old_cls = old_name.rsplit(".", 1)
        new_mod, new_cls = new_name.rsplit(".", 1)
        replace_multi_suffix(
            lines,
            (f".. {direc}:: {old_cls}", f"   :module: {old_mod}"),
            (f".. {direc}:: {new_cls}", f"   :module: {new_mod}"),
        )


def replace_multi_suffix(
    lines: StringList, old: tuple[str, str], new: tuple[str, str]
) -> None:
    if len(old) != len(new) != 2:  # noqa: PLR2004  # pragma: no cover
        msg = "Only supports replacing 2 lines"
        raise NotImplementedError(msg)
    for l, line in enumerate(lines):
        start = line.find(old[0])
        if start == -1:
            continue
        prefix = line[:start]
        suffix = line[start + len(old[0]) :]
        if lines[l + 1].startswith(prefix + old[1]):
            break
    else:
        return
    lines[l + 0] = prefix + new[0] + suffix
    lines[l + 1] = prefix + new[1]


def setup(app: Sphinx) -> None:
    # no support for new autodoc
    if getattr(app.config, "autodoc_use_legacy_class_based", True):
        ClassDocumenter.add_directive_header = add_directive_header  # type: ignore[method-assign]
